@@ -3,7 +3,7 @@ This file is part of Swapspace.
 
 Copyright (C) 2005,2006, Software Industry Industry Promotion Agency (SIPA)
 Copyright (C) 2010, Jeroen T. Vermeulen
-Copyright (C) 2019, Jacob Adams
+Copyright (C) 2019,2021, Jacob Adams
 Written by Jeroen T. Vermeulen.
 
 Swapspace is free software; you can redistribute it and/or modify
@@ -55,7 +55,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #define O_LARGEFILE 0
 #endif
 
-// Don't follow symlinks, if possible; they may pose a security risk 
+// Don't follow symlinks, if possible; they may pose a security risk
 #ifndef O_NOFOLLOW
 #define O_NOFOLLOW 0
 #endif
@@ -157,8 +157,6 @@ bool to_swapdir(void)
     return false;
   }
 
-  swappath_len = strlen(swappath);
-
 #ifndef NO_CONFIG
   // Get rid of any "/./", "//", and "/../" clutter that might be in swappath.
   // This is needed because we want to recognize our swapfiles in /proc/swaps,
@@ -169,6 +167,8 @@ bool to_swapdir(void)
     return false;
   }
 
+  swappath_len = strlen(swappath);
+
   // Remove trailing slash, if any
   if (swappath[swappath_len-1] == '/')
   {
@@ -180,28 +180,86 @@ bool to_swapdir(void)
     logm(LOG_ERR, "Not supported: swap path contains whitespace");
     return false;
   }
+#else
+  swappath_len = strlen(swappath);
 #endif
+  return swapdir_config();
+}
+
+
+/// Configure permissions and attributes correctly on the swap directory
+bool swapdir_config(void) {
+  int fd = open(swappath, O_RDONLY);
+  if (fd < 0) {
+    log_perr(LOG_ERR, "Unable to open swap directory", errno);
+	return false;
+  }
   struct stat swapdirstat;
   mode_t mode = S_IFDIR | S_IRUSR | S_IWUSR | S_IXUSR;
-  if (stat(swappath, &swapdirstat) < 0)
+  if (fstat(fd, &swapdirstat) < 0)
   {
-    log_perr(LOG_ERR, "Unable to stat swap path", errno);
+    log_perr(LOG_ERR, "Unable to stat swap directory", errno);
     return false;
   }
   if (swapdirstat.st_mode != mode)
   {
-    logm(LOG_WARNING, "Swap file directory permissions wrong, fixing");
+    logm(LOG_WARNING, "Swap directory permissions wrong, fixing");
     if (verbose)
     {
       logm(LOG_DEBUG, "Actual: %o\nExpected: %o", swapdirstat.st_mode, mode);
     }
     if (chmod(swappath, mode) < 0)
     {
-      log_perr(LOG_ERR, "Unable to change swap file directory permissions", errno);
+      log_perr(LOG_ERR, "Unable to change swap directory permissions", errno);
       return false;
     }
   }
 
+  int attr;
+  int err = ioctl(fd, FS_IOC_GETFLAGS, &attr);
+  if (err != 0)
+  {
+    // Don't worry about attributes if they're not supported
+    if(errno != ENOTSUP)
+    {
+      log_perr(LOG_NOTICE, "Unable to get swap path attributes, "
+			"may not be able to create swap files", errno);
+    }
+    return true;
+  }
+
+
+  // Unset compression, if set
+  if (attr & FS_COMPR_FL)
+  {
+    if (verbose)
+    {
+      logm(LOG_NOTICE, "Compression set on swap directory, fixing");
+    }
+    attr &= ~FS_COMPR_FL;
+    err = ioctl(fd, FS_IOC_SETFLAGS, &attr);
+    if (err != 0 && errno != ENOTSUP)
+    {
+      log_perr(LOG_NOTICE, "Unable to unset compression, "
+			  "may not be able to create swap files", errno);
+    }
+  }
+
+  // Ensure NOCOW is set, though only needed for btrfs
+  if (!(attr & FS_NOCOW_FL))
+  {
+    if (verbose)
+    {
+      logm(LOG_NOTICE, "NOCOW not set on swap directory, fixing");
+    }
+    attr |= FS_NOCOW_FL;
+    err = ioctl(fd, FS_IOC_SETFLAGS, &attr);
+    if (err != 0 && errno != ENOTSUP)
+    {
+      log_perr(LOG_NOTICE, "Unable to set NOCOW, "
+			  "may not be able to create swap files", errno);
+    }
+  }
   return true;
 }
 
@@ -235,14 +293,6 @@ static bool proc_swaps_read_ok = false;
 /// Can we allocate swapfiles using posix_allocate on this filesystem?
 static bool pfalloc_ok = true;
 
-/* Store filesystem type so we don't check everytime we allocate a swapfile
- * __fsword_t is a glibc-internal type, but the docs claim unsigned int is fine
- */
-static unsigned int fstype = 0;
-
-/// Does the filesystem support the NOCOW attribute?
-static bool nocow_ok = true;
-
 /// Print status information to stdout
 void dump_stats(void)
 {
@@ -433,70 +483,11 @@ static memsize_t fill_swapfile(const char file[], int fd, memsize_t size)
   return bytes;
 }
 
-/// Ensure COW updates don't apply to swapfiles.
-/* Only needed to BTRFS and other COW filesystems.
- * But can't hurt to apply everywhere.
- */
-int set_no_cow(int fd)
-{
-  int attr;
-  int err;
-
-  err = ioctl(fd, FS_IOC_GETFLAGS, &attr);
-  if (err != 0)
-  {
-    if (errno == ENOTSUP)
-    {
-      nocow_ok = false;
-      logm(LOG_NOTICE, "Filesystem does not support attributes!");
-    }
-    else return err;
-  }
-  attr |= FS_NOCOW_FL;
-  err = ioctl(fd, FS_IOC_SETFLAGS, &attr);
-  if (err != 0)
-  {
-    if (errno == ENOTSUP)
-    {
-      nocow_ok = false;
-      logm(LOG_NOTICE, "NOCOW attribute not supported!");
-    }
-    else return err;
-  }
-  return 0;
-}
-
-
-/// Handle filesystem-specific operations needed for swapfiles
-/* Some filesystems, like BTRFS, require special operations to be performed on
- * files before they can be used as swapfiles
- * Clobbers localbuf if special operations are performed
- */
-void specialfs(const char path[])
-{
-  int err;
-  /* All swapfiles exist on the same filesystem, so we should only need to
-   * call this once
-   */
-  if (fstype == 0)
-  {
-    struct statfs buf;
-    err = statfs(path, &buf);
-    if (err == 0) fstype = buf.f_type;
-    else log_perr_str(LOG_WARNING, "Could not detect filesystem", path, errno);
-  }
-  if (fstype == BTRFS_SUPER_MAGIC)
-  {
-    err = runcommandformat("%s property set '%s' compression none", "btrfs", path);
-    if (err != 0) logm(LOG_WARNING, "Could not disable BTRFS compression! Return Code: %d", err);
-  }
-}
-
 /// Create file to be used as swap.  Clobbers localbuf.
 /**
  * @param filename File to be created
  * @param size Desired size in bytes (but already rounded to page size)
- * @return Size of new swapfile, which may differ from requested size.  Zero 
+ * @return Size of new swapfile, which may differ from requested size.  Zero
  * indicates failure, in which case the file is deleted.
  */
 static memsize_t make_swapfile(const char file[], memsize_t size)
@@ -519,21 +510,13 @@ static memsize_t make_swapfile(const char file[], memsize_t size)
     return 0;
   }
 
-  if (nocow_ok)
-  {
-    int err = set_no_cow(fd);
-    if (unlikely(err != 0)) log_perr_str(LOG_WARNING, "Could not disable COW", file, errno);
-  }
-
-  specialfs(file);
-
   if (unlikely(!fill_swapfile(file, fd, size)))
   {
     size = 0;
     unlink(file);
   }
   close(fd);
-  
+
   return size;
 }
 
@@ -595,7 +578,7 @@ bool activate_old_swaps(void)
 #ifndef NO_CONFIG
       if (!quiet) logm(LOG_INFO, "Found old swapfile '%d'", seqno);
 #endif
-      const memsize_t size = filesize(d->d_name); 
+      const memsize_t size = filesize(d->d_name);
       if (likely(size > min_swapsize && enable_swapfile(d->d_name)))
       {
 	swapfiles[seqno].size = size;
